@@ -111,15 +111,7 @@ class Nami():
         """Get all information for a single instance (status + GPU info)."""
         config = self.config["instances"][name]
         # Fetch GPU info first; this SSH call is sufficient to decide if the host is reachable.
-        gpu_info_lines = self.get_gpu_info(name)
-
-        # Determine online/offline status from the first returned line.
-        first_line = gpu_info_lines[0] if gpu_info_lines else "❌ Error"
-        if first_line.startswith("❌"):
-            status = "❌ Offline"
-        else:
-            status = "✅ Online"
-
+        gpu_info_lines, status = self.get_gpu_info(name)
         return name, config, status, gpu_info_lines
 
     def list_instances(self):
@@ -169,9 +161,10 @@ class Nami():
                 print(f"   Command: ssh {config['user']}@{config['host']} {local_port}")
             if config.get('description'):
                 print(f"   Description: {config['description']}")
-            print("   GPUs:")
-            for gpu_line in gpu_info_lines:
-                print(gpu_line)
+            if gpu_info_lines is not None:
+                print("   GPUs:")
+                for gpu_line in gpu_info_lines:
+                    print('     ' + gpu_line)
             print()
 
     def get_template(self, template_name):
@@ -260,76 +253,82 @@ class Nami():
             When ``True`` the instance's ``local_port`` value is forwarded via
             ``ssh -L``.  By default no port forwarding is performed.
         """
-        with Connection(instance_name, self.config, enable_port_forwarding=forward) as ssh:
+        with Connection(instance_name, self.config, enable_port_forwarding=forward, personal_config=self.personal_config) as ssh:
             ssh.run(command)
 
-    def connect_ssh(self, instance_name, command=None, forward=False):
+    def connect_ssh(self, instance_name, forward=False):
         """Open an interactive SSH session to *instance_name*.
 
         If *forward* is ``True`` the configured ``local_port`` will be
         forwarded.
         """
-        with Connection(instance_name, self.config, enable_port_forwarding=forward) as ssh:
-            ssh.run_interactive(command)
+        with Connection(instance_name, self.config, enable_port_forwarding=forward, personal_config=self.personal_config) as ssh:
+            ssh.run_interactive()
 
     def get_gpu_info(self, name):
         """Get GPU information for an instance."""
         if name not in self.config.get("instances", {}):
             return ["❌ Not configured"]
         
-        config = self.config["instances"][name]
         try:
-            # Run nvidia-smi to get GPU information
-            ssh_cmd = [
-                "ssh", "-o", "ConnectTimeout=5", "-o", "BatchMode=yes",
-            ]
-            if config.get('port') is not None:
-                ssh_cmd.append(f"-p{config['port']}")
-            ssh_cmd.extend([
-                f"{config['user']}@{config['host']}",
-                "nvidia-smi --query-gpu=index,name,utilization.gpu,memory.used,memory.total --format=csv,noheader,nounits 2>/dev/null || echo 'NO_GPU'"
-            ])
-            result = subprocess.run(ssh_cmd, capture_output=True, text=True, timeout=10)
-            
-            if result.returncode != 0:
-                return ["❌ SSH Failed"]
-            
-            output = result.stdout.strip()
-            if not output or output == "NO_GPU":
-                return ["🔘 No GPU"]
-            
-            # Parse GPU information
-            gpu_lines = output.split('\n')
-            gpu_info = []
-            for line in gpu_lines:
-                if line.strip():
-                    parts = [p.strip() for p in line.split(',')]
-                    if len(parts) >= 5:
-                        gpu_idx, gpu_name, gpu_util, mem_used, mem_total = parts[:5]
-                        try:
-                            gpu_util = int(gpu_util)
-                            mem_used = int(mem_used)
-                            mem_total = int(mem_total)
-                            mem_percent = int((mem_used / mem_total) * 100) if mem_total > 0 else 0
-                            
-                            # Color code based on utilisation
-                            if gpu_util >= 50:
-                                util_color = "🔴"
-                            elif mem_percent >= 50:
-                                util_color = "🟠"
-                            elif gpu_util >= 10:
-                                util_color = "🟡"
-                            else:
-                                util_color = "🟢"
-                            
-                            gpu_info.append(f"     {util_color} GPU{gpu_idx}: {gpu_util:3d}% | Mem: {mem_percent:3d}% | {gpu_name}")
-                        except (ValueError, ZeroDivisionError):
-                            gpu_info.append(f"     🔘 GPU{gpu_idx}: Error parsing")
-            
-            return gpu_info or ["🔘 No GPU data"]
+            with Connection(name, self.config, personal_config=self.personal_config) as conn:
+                result = conn.run(
+                    "nvidia-smi --query-gpu=index,name,utilization.gpu,memory.used,memory.total --format=csv,noheader,nounits 2>/dev/null || echo 'NO_GPU'", 
+                    capture=True
+                )
+                output = result.stdout.strip()
+                if not output or output == "NO_GPU":
+                    return ["🔘 No GPU"], "✅ Online"
+                
+                # Parse GPU information
+                gpu_lines = output.split('\n')
+                gpu_info = []
+                for line in gpu_lines:
+                    if line.strip():
+                        parts = [p.strip() for p in line.split(',')]
+                        if len(parts) >= 5:
+                            gpu_idx, gpu_name, gpu_util, mem_used, mem_total = parts[:5]
+                            try:
+                                gpu_util = int(gpu_util)
+                                mem_used = int(mem_used)
+                                mem_total = int(mem_total)
+                                mem_percent = int((mem_used / mem_total) * 100) if mem_total > 0 else 0
+                                
+                                # Color code based on utilisation
+                                if gpu_util >= 50:
+                                    util_color = "🔴"
+                                elif mem_percent >= 50:
+                                    util_color = "🟠"
+                                elif gpu_util >= 10:
+                                    util_color = "🟡"
+                                else:
+                                    util_color = "🟢"
+                                
+                                gpu_info.append(f"{util_color} GPU{gpu_idx}: {gpu_util:3d}% | Mem: {mem_percent:3d}% | {gpu_name}")
+                            except (ValueError, ZeroDivisionError):
+                                gpu_info.append(f"🔘 GPU{gpu_idx}: Error parsing")
+                
+                return gpu_info or ["🔘 No GPU data"], "✅ Online"
         
-        except Exception:
-            return ["❌ Error"]
+        except Exception as e:
+            error_str = str(e).lower()
+            # Extract output from the error message if present
+            if "output:" in error_str:
+                _, output_part = error_str.split("output:", 1)
+                output_str = output_part.strip().lower()
+            else:
+                output_str = ""
+
+            full_error = error_str + " " + output_str
+
+            if any(keyword in full_error for keyword in ['connection refused', 'connection rejected', 'connection denied']):
+                return None, "⚠️  Unavailable"
+            elif any(keyword in full_error for keyword in ['unreachable', 'no route']):
+                return None, "❌ Network Error"
+            elif any(keyword in full_error for keyword in ['timeout', 'timed out']):
+                return None, "⏱️  Timeout"
+            else:
+                return None, "❌ Error: " + str(e)
 
     def _add_key_to_instance(self, name, key_file):
         config = self.config.get("instances", {}).get(name)
@@ -456,6 +455,17 @@ def main():
     transfer_parser.add_argument("--rsync_opts", default="-avz --progress", help="Extra rsync options")
     transfer_parser.add_argument("--endpoint", help="Custom S3 endpoint URL")
 
+    # NFS mesh mounting command
+    nfs_parser = subparsers.add_parser("nfs", help="NFS mesh mounting operations")
+    nfs_subparsers = nfs_parser.add_subparsers(dest="nfs_action", required=True, help="NFS actions")
+
+    mesh_parser = nfs_subparsers.add_parser("mount-mesh", help="Setup and mount full NFS mesh among instances")
+    mesh_parser.add_argument("--instances", nargs="+", help="Instance names to include in the mesh")
+    mesh_parser.add_argument("--export_dir", default="/", help="Path to export on each server")
+    mesh_parser.add_argument("--mount_base", default="/mnt/peers", help="Base directory on clients for peer mounts")
+    mesh_parser.add_argument("--nfs_version", default="4", help="NFS protocol version for mounts")
+    mesh_parser.add_argument("--max_workers", type=int, default=24, help="Concurrency for operations")
+
     # Download from S3
     from_s3_parser = subparsers.add_parser("from_s3", help="Download files/folders from S3 to an instance")
     from_s3_parser.add_argument("--dest_instance", required=True)
@@ -524,52 +534,69 @@ def main():
         else:
             print(f"❌ Unknown ssh-key action: {args.ssh_key_action}. Available: add")
     elif args.command == "transfer":
-        dest_path = args.dest_path or args.source_path
         if args.method == "rsync":
             rsync.transfer_via_rsync(
                 source_instance=args.source_instance,
                 dest_instance=args.dest_instance,
                 source_path=args.source_path,
-                dest_path=dest_path,
+                dest_path=args.dest_path or args.source_path,
                 exclude=args.exclude_patterns or "",
                 rsync_opts=args.rsync_opts,
                 archive=args.archive,
                 config=vm.config,
+                personal_config=vm.personal_config
             )
         elif args.method == "s3":
             s3.transfer_via_s3(
                 source_instance=args.source_instance,
                 dest_instance=args.dest_instance,
                 source_path=args.source_path,
-                dest_path=dest_path,
+                dest_path=args.dest_path or args.source_path,
                 s3_bucket=vm.personal_config["s3_bucket"],
                 aws_profile=vm.personal_config.get("aws_profile", "default"),
                 exclude=args.exclude_patterns or "",
                 archive=args.archive,
                 endpoint=args.endpoint,
                 config=vm.config,
+                personal_config=vm.personal_config
             )
+    elif args.command == "nfs":
+        if args.nfs_action == "mount-mesh":
+            from .nfs.nfs import setup_and_mount_full_mesh
+            setup_and_mount_full_mesh(
+                instances=args.instances,
+                export_dir=args.export_dir,
+                mount_base=args.mount_base,
+                nfs_version=args.nfs_version,
+                max_workers=args.max_workers,
+                config=vm.config,
+                personal_config=vm.personal_config
+            )
+        else:
+            print(f"❌ Unknown nfs action: {args.nfs_action}. Available: mount-mesh")
     elif args.command == "from_s3":
         s3.download_from_s3(
             dest_instance=args.dest_instance,
             source_path=args.source_path,
             dest_path=args.dest_path,
-            aws_profile=args.aws_profile or vm.personal_config.get("aws_profile", "default"),
+            aws_profile=vm.personal_config.get("aws_profile", "default"),
             exclude=args.exclude_patterns or "",
             archive=args.archive,
             endpoint=args.endpoint,
             config=vm.config,
+            personal_config=vm.personal_config
         )
     elif args.command == "to_s3":
         s3.upload_to_s3(
             source_instance=args.source_instance,
             source_path=args.source_path,
             dest_path=args.dest_path,
-            aws_profile=args.aws_profile or vm.personal_config.get("aws_profile", "default"),
+            aws_profile=vm.personal_config.get("aws_profile", "default"),
             exclude=args.exclude_patterns or "",
             archive=args.archive,
             endpoint=args.endpoint,
             config=vm.config,
+            personal_config=vm.personal_config
         )
     elif args.command == "template":
         template_vars: dict[str, str] = {}
